 !
 !   This file is part of ????.
 !
 !   pls4snp is free software; you can redistribute it and/or modify
 !   it under the terms of the GNU General Public License as published by
 !   the Free Software Foundation; either version 2 of the License, or
 !   (at your option) any later version.
 !
 !   pls4snp is distributed in the hope that it will be useful,
 !   but WITHOUT ANY WARRANTY; without even the implied warranty of
 !   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 !   GNU General Public License for more details.
 !
 !   You should have received a copy of the GNU General Public License
 !   along with pls4snp; if not, write to the Free Software
 !   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 !
 !   Copyright 2013 Olivier Filangi
 !
 !
! ============================================================================
! Name        : BuildMqqFromBlup.f03
! Author      : Olivier Filangi
! Version     :
! Copyright   : GPLv3
! Description :
! ============================================================================

program gwas_blupf90_muller
    use kinds
    use ilapack
    use const
    use io         ! input/output to read and write data
    use cli        ! command line interface
    use omp_lib    ! OMP support
    use blupf90    ! module Blupf90 Misztal/Tsuruta/Aguilar
    use out_scriptR ! generating R graph
    use mt_stream   ! mersenne twister PRNG

    implicit none

    !Arguments de l'application
    character(len=len_long_cli)   :: genotypeFile=""    ! Fichier contenant les genotype au format AIPL
    character(len=len_long_cli)   :: mapFile=""         ! Fichier decrivant les groupes de liason
    character(len=len_long_cli)   :: pedigFile=""       ! Pedigree contenant les individus numeroté de 0 à N
    character(len=len_long_cli)   :: modelFile=""       ! Fichier de description des performances et des effets de nuisances
    character(len=len_long_cli)   :: dataFile=""        ! Fichier des performances et effets fixé et covariable
    character(len=len_long_cli)   :: solutionFile=""    ! Fichier solution du GWAS
    character(len=len_long_cli)   :: outFile=""         ! Fichier resultat
    character(len=len_long_cli)   :: distributionFile=""! Fichier sortie des simuls des Psmins

    integer                       :: npostotal          ! nombre total de snp testé
    integer                       :: npos = 0           ! Nombre reel de position testé (MAF correcte)
    character(len=20) ,dimension(:),pointer   :: chr  => null()   ! Nom des groupe de Liaison
    integer ,dimension(:),pointer :: startpos => null()          ! premier snp testé
    integer ,dimension(:),pointer :: endpos  => null()           ! dernier snp testé
    integer ,dimension(:),pointer :: nmark  => null()            ! nombre de marqueurs pour chaque groupe de liaison
    integer                       :: nchr = 0           ! number of chromosome/GL to test
    integer                       :: nu = 0 ! number of animal genotyped

    integer                                        :: idread
    ! correspondance id à une renumerotation pour acceder aux structure de donnees
    integer           ,dimension(:) ,pointer       :: id   => null()           ! correspondance index animaux , numero identifiant
    integer           ,dimension(:) ,pointer       :: revid_global => null()   ! correspondance identifiant animaux , index du fichier des genotypes
    integer                                        :: out_unit                 ! unit pour la sortie à l'ecran ou dans un fichier

    !status des animaux genotype
    integer , dimension(:,:)    , pointer          :: genotype => null()

    integer                                        :: nbe ! number of fixed effect

    ! to read entries from blupf90
    integer            , dimension(:), allocatable :: idmax,ivec,jvec
    real(kind=dp)      , dimension(:), allocatable :: vvec

    ! to build matrix from entries Blupf90
    real(kind=dp)      , dimension(:,:), allocatable :: A
    real(kind=dp)      , dimension(:), allocatable   :: B
    real(kind=dp)      , dimension(:,:), allocatable :: C
    real(kind=dp)      , dimension(:,:), allocatable :: D
    real(kind=dp)      , dimension(:)  , allocatable :: Eq1,Eq2

    ! name of matrix file blupf90
    character(len=len_long_cli)                      :: defstr

    real(kind=dp)      , dimension(:,:), pointer     :: M => null() ,P => null()
    real(kind=dp)      , dimension(:)  , allocatable :: SIGMA,Pdiag
    real(kind=dp)      , dimension(:)  , allocatable :: U,Ts,Psmin,Pvalues ! Simulation structure

    integer :: ios,i,q1,q2,ndim,isim,s,ifail,idth
    character(len=10) :: qstr
    character(len=1)  :: TAIL_G01ECF = 'U',JOBZ,UPLO
    real(kind=dp)     :: Fq1q2,vare,invvare,vtemp,vary,varg,h2,counti
    integer           :: info,j,iani,imark,startblup

    real                             :: start,finish

    logical                          :: sort,calcul_vary
    character(len=len_long_cli)      :: option_l
    character(len=1)                 :: option_s_opt
    real(kind=dp)                    :: meany

    ! New implementation - Calling blupf90 by subroutines
    type(Param_blupf90)              :: paramblupf90   ! parametre d entre du blupf90
    type(Solution_blupf90)           :: solblupf90     ! solution en sortie de procedure blupf90

    integer, dimension(:,:),pointer  :: data_ped  => null() ! pedigree structure
    integer                          :: ndata_ped ! number of record for pedigree structure
    real, dimension(:,:),pointer     :: data => null() ,newdata => null()
    integer                          :: ndata,ncol,nnewdata
    real,dimension(:,:),allocatable  :: guser
    integer,dimension(:),allocatable :: revid_currenthaplo
    integer                          :: nguser,perc,percsave,pvalueNaN,ich
    logical,dimension(:),allocatable :: okmarker
    real,dimension(:)  ,allocatable  :: mafinfo

    !Solution GWAS
    integer,dimension(:),allocatable :: performer
    real   ,dimension(:),allocatable :: test
    real   ,dimension(:),allocatable :: pvaluespos
    real   ,dimension(:),allocatable :: sol
    real   ,dimension(:),allocatable :: solmu
    real   ,dimension(:),allocatable :: std
    real   ,dimension(:),allocatable :: stdmu

    real   ,dimension(:,:),allocatable :: maxpvalues
    integer,dimension(:,:),allocatable :: maxpvalues_index
    real                             :: pvallog10
    integer , parameter              :: u_sol=2765 ! unit for solution file

    type(mt_state) , allocatable ,dimension(:) :: mts_stream
    integer        , allocatable ,dimension(:) :: iseed
    real(kind=dp)                              :: rd1,rd2

    call cpu_time(start)
    !====================================================================================
    !
    !     Arguments
    !
    !====================================================================================
    write (*,*)
    write (*,*) "==============  OPTIONS ================"
    write (*,*) "--geno,      -g     : genotype file"
    write (*,*) "--model,     -z     : Model file"
    write (*,*) "--pedig,     -p     : Pedigree file"
    write (*,*) "--data,      -d     : Data file"
    write (*,*) "--map,       -c     : map file"
    write (*,*) "--solution,  -b     : Solution file"
    write (*,*) "--out,       -o     : Result file"
    write (*,*) "--out-dist,  -y     : Distribution Psmin file"
    write (*,*) "--idlength,  -i     : Length ID in the genotype file (default=10)"
    write (*,*) "--vary,      -v     : phenotypic variance ",&
     "(optional, by default this quantity is compute with the classical estimator)"
    write (*,*) "--h2,        -h     : heritability"
    write (*,*) "--nsim,      -s     : Number of simulation (default=1000)"
    write (*,*) "--maf,       -m     : Minimum MAF (default=0.02)"
    write (*,*) "--nmaxpval,  -x     : Maximum number of pvalue to rank (default=30)."
    write (*,*) "--nthread,   -N     : Number of thread usable by the program (default=1)."
    write (*,*)

    !==================================================================================================================
    !
    ! Lecture des options en entrée de l'application
    !
    ! Initialisations des parametres suivants
    !
    ! genotypeFile   : fichier contenant les genotypes
    ! modelFile      : fichier decrivant le fichier de phenotypes et des effets de nuisance
    ! pedigFile      : fichier pedigree
    ! dataFile       : fichier des phenotypes et effets fixé
    ! sizeid         : nombre de caracteres du codage de l'ID dans le fichier de genotype
    ! startpos       : premiere position à testé dans le GWAS
    ! endpos         : derniere position
    ! npostotal      : nombre de position total
    ! vary           : variance phenotypique
    ! h2             : heritabilité
    ! vare           : variance residuelle déduit de h2 et vary
    ! varg           : variance genetique deduit de h2 et vary
    ! invvare        : inverse de la variance residuelle
    ! minmaf         : minimum MAF autorisé pour l'analyse GWAS
    ! nsim
    !
    !
    call set_options_gwas

    call set_map(mapFile,chr,startpos,endpos,nmark,nchr)

    write (out_unit,*) "Number of Linkage Groupe:",nchr
    write (out_unit,*)
    write (out_unit,fmt="(a11,a8,a8,a8)") "Chr","Start","End","Number"
    do ich=1,nchr
     write (out_unit,fmt="(a10,1x,I7,1x,I7,1x,I7)") chr(ich),startpos(ich),endpos(ich),nmark(ich)
    end do
    write (out_unit,*)
    npostotal = maxval(endpos(:nchr))
    startblup = minval(startpos(:nchr)) !ca ne sert à rien de faire les blup en dessous de ce marqueurs, il ne sont pas dans un GL
    !==================================================================================================================
    !
    ! Lecture du pseudo fichier parametre (description du fichier des performance et des effets de nuisances)
    !
    ! Construction d'une variable de type type(Param_blupf90)
    ! Cette structure remplace le fichier parametre donnee initialement au blupf90
    !
    call fill_structure_data_blupf90(dataFile,modelFile,paramblupf90,ndata,ncol,data)
    if ( calcul_vary ) then
      meany = sum(data(:,5)) / real(ndata)
      vary = 0.d0
      do i=1,ndata
       vary = vary + (data(i,5)-meany)*(data(i,5)-meany)
      end do
      vary = vary / (real(ndata) - 1)
    end if
    write (out_unit,*) "[phenotypic variance=",vary,"]"
    call fill_structure_pedigree_blupf90(pedigFile,paramblupf90,ndata_ped,data_ped)
    vare = (1.d0 - h2) * vary
    write (out_unit,*) "[residual variance=",vare,"]"
    invvare = 1.d0 / vare
    varg= h2*vary
    write (out_unit,*) "[genetic variance=",varg,"]"

    call set_covariance_genetique_blupf90(paramblupf90,varg)
    call set_residuelle_blupf90(paramblupf90,vare)
    call set_general_blupf90(paramblupf90)
    if (verbose) &
       call print_information_model_blupf90(paramblupf90)

    ! on utilise pas cette structure de données mais blupf90 en a besoin...
    nguser=0
    allocate (guser(nguser,3))
    guser=0.d0

    !==================================================================================================================
    !
    ! Lecture du fichier des genotypes
    !
    call load_genotype(genotypeFile,npostotal,id,genotype,nu)

    print *,"read ",nu," genotype."
    print *,"Genotype First ID [",id(1),"] =>",genotype(1,:min(10,npostotal)),"...",genotype(1,max(npos-10,npostotal):)
    print *,"Genotype Last ID [",id(nu),"] =>",genotype(nu,:min(10,npostotal)),"...",genotype(nu,max(npos-10,npostotal):)

    if ( maxval(id)> 900000) then
      write (0,*) "max ID animal in genotype file seems to big = ",maxval(id)
      stop
    end if
    !Tableau pour retrouver les indices des animaux gere en interne
    allocate (revid_global(maxval(id)))
    revid_global=-1
    do i=1,nu
        revid_global(id(i))=i
    end do

    !Les marqueurs valides pour l'analyse
    allocate (okmarker(startblup:npostotal),mafinfo(startblup:npostotal))

    !On utilise un tableau de boolean pour gerer seulement les marqueurs au dessus d'une MAF donnée par l'utilisateur
    call remove_maf(nu,startblup,npostotal,genotype,okmarker,minmaf,mafinfo)
    !====================================================================================
    !
    ! GWAS : blup pour chaque position
    !
    !====================================================================================
    ! nombre de marker à traiter reellement
    npos = count(okmarker)
    ! addition du nombre de niveau des effets fixés
    nbe = sum(paramblupf90%nlev(:paramblupf90%neff-2))

    allocate (A(npos,nbe),B(npos),C(npos,nbe),D(npos,nu))
    allocate (performer(npos))
    allocate (test(npos))
    allocate (pvaluespos(npos))
    allocate (sol(npos))
    allocate (solmu(npos))
    allocate (std(npos))
    allocate (stdmu(npos))

    write (*,*) "============================"
    write (*,*) "call ",npos," BLUPF90."
    counti=0

    !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(newdata,revid_currenthaplo,imark,nnewdata,solblupf90)
    ! newdata correspiond au fichier des donnees traité par blupf90
    ! l'avant derniere colonne correspond au status du marqueur
    allocate (newdata(ndata,ncol))
    ! revid_currenthaplo est le tableau des correspondance pour retrouver les ID du fichier de genotypage et des individus phenotypés
    ! attention, dans new data on retrouve les individus phénotypés et genotypés pour ce marqueur !
    ! Il se peut qu'on ne retrouve pas certain individus pour 2 marqueurs differentes si le genotypes est manquant !
    allocate (revid_currenthaplo(ndata))
    percsave=-1
    ! Remplissage des structures A,B,C,D
    !$OMP DO SCHEDULE(DYNAMIC)
    do i=startblup,npostotal
        !$OMP ATOMIC
        counti=counti+1
        !Affichage dfe l'etat du calcul à la charge du thread 0
        if ( omp_get_thread_num() == 0 ) then
            perc = nint((real(counti)/(real((npostotal-startblup+1))))*100)
            if (perc /= percsave ) then
             write (*,fmt='(a,"===>",(" [",i3,"%] "))',advance='no') &
                achar(13),perc
             percsave=perc
            end if
        end if

        ! On change la colonne correspondant au status du SNP
        ! newdata a une dimension ligne = ndata mais contient en sortie nnewdata ligne effectives (animaux à la fois phenotype et genotype pour le marqueur)
        if (.not. okmarker(i)) then
            cycle
        end if

        imark = count(okmarker(:i))
        call set_marker_blupf90(revid_global,nu,npostotal,genotype,ndata,ncol,data,nnewdata,newdata,i,revid_currenthaplo)
        call exec_blupf90(paramblupf90,ndata,nnewdata,ncol,newdata,ndata_ped,data_ped,nguser,guser,solblupf90,.false.)

        A(imark,:) = solblupf90%A(1,:)
        B(imark) = solblupf90%B(1,1)
        C(imark,:) = solblupf90%C(1,:)
        D(imark,:) = 0.d0
        do iani=1,solblupf90%nani
            if ( solblupf90%Dani(iani) == 0 ) cycle ! valeur =0
            D(imark,revid_global(solblupf90%Dani(iani))) = solblupf90%D(1,iani)
        end do

        performer(imark) = solblupf90%nani
        test(imark) = solblupf90%test
        pvaluespos(imark) = solblupf90%pvalue
        sol(imark)     = solblupf90%sol(1) ! on ne gere pas les haplotypes encore
        solmu(imark)   = solblupf90%solmu
        std(imark)     = solblupf90%std(1) ! on ne gere pas les haplotypes encore
        stdmu(imark)   = solblupf90%stdmu

        call release_solutionblupf90(solblupf90)
    end do
    !$OMP END DO
    deallocate (newdata,revid_currenthaplo)
    !$OMP END PARALLEL

    write (*,*)
    write (*,*) "Writing ",trim(solutionFile)," file [id snp chr maf perfo test pvalue sol solmu std stdmu]."

    ! Ecriture du fichier solution
    open(u_sol,file=solutionFile,action="write",iostat=ios)
    if ( ios /= 0 ) then
        write (*,*) "can not open solution file."
        stop
    end if

    allocate (maxpvalues_index(nchr,MAXPVALUESPRINT),maxpvalues(nchr,MAXPVALUESPRINT))
    maxpvalues_index = 0
    maxpvalues = -9999.d0
    ! * header *
    write (u_sol,fmt="(a8,a9,a9,a8,a8,a16,a16,a16,a16,a16,a16)") &
    "id", "snp", "chr" ,"maf", "nbani", "test", "pvalue", "sol", "solmu", "std", "stdmu"

    imark=0
    pvalueNaN = 0

    do ich=1,nchr
     do i=startpos(ich),endpos(ich)
        if ( okmarker(i) ) then
            imark=count(okmarker(:i))
            pvallog10 = -log10(pvaluespos(imark))
            ! si l'un des plus gros effet alors on enregistre la position (pour de l'affichage)
            if (pvallog10>maxpvalues(ich,size(maxpvalues,2))) then
              do j=1,size(maxpvalues,2)
               if ( pvallog10 > maxpvalues(ich,j)) exit
              end do
              maxpvalues(ich,j+1:) = maxpvalues(ich,j:size(maxpvalues,2)-1)
              maxpvalues_index(ich,j+1:) = maxpvalues_index(ich,j:size(maxpvalues,2)-1)
              maxpvalues(ich,j) = pvallog10
              maxpvalues_index(ich,j) = i
            end if
            write (u_sol,fmt="(I7,1x,I7,2x,a7,2x,F6.3,1x,I7,1x,F15.8,2x,F15.8,2x,F15.8,2x,F15.8,2x,F15.8,2x,F15.8)") &
                imark, i, trim(chr(ich)), mafinfo(i) , performer(imark),test(imark),&
                pvaluespos(imark),sol(imark),solmu(imark),std(imark),stdmu(imark)
            if ( pvaluespos(imark) /= pvaluespos(imark) ) then
              pvalueNaN = pvalueNaN + 1
            end if
        end if
     end do
    end do

    close(u_sol)
    deallocate (performer,test,pvaluespos,sol,solmu,std,stdmu)

    !====================================================================================
    !
    ! Construction des Mqq / CHR
    !
    !====================================================================================

    do ich=1,nchr
    npos=nmark(ich)
    print *," =============== Prepare covariance matrix to compute rejection thresholds (",&
           trim(chr(ich)),")============== "
    ndim = nbe+1+nbe+nu+1 ! number of entries in file from blupf90 + 1 to manage last read
    allocate (M(npos,npos))
    counti=0
    percsave=-1
    write (*,*) "==="
    !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(idmax,ivec,jvec,vvec,Eq1,Eq2,Fq1q2,qstr,q2)
    allocate (idmax(ndim),ivec(ndim),jvec(ndim),vvec(ndim),Eq1(nu),Eq2(nu))
    !$OMP DO SCHEDULE(DYNAMIC)
    do q1=1,npos
        ! On renumerote de 1 à 3 si l'allele 0 est representé dans la population le SNP q
        if ( count(genotype(:,q1) == 0 ) > 0 ) then
            where (genotype(:,q1) >=0 .and. genotype(:,q1) <=2 )
                Eq1 = genotype(:,q1)+1
            elsewhere
                Eq1 = 0.d0
            end where
        else
            Eq1 = genotype(:,q1)
            where ( Eq1 > 2 .or. Eq1 < 0 )
                Eq1 = 0.d0
            end where
        end if

        do q2=q1+1,npos
            !$OMP ATOMIC
            counti=counti+1
            if ( omp_get_thread_num() == 0 ) then
             perc = nint((real(counti)/(real((npos*npos)/2)))*100)
             if (perc /= percsave ) then
              write (*,fmt='(a,"===>",(" [",i3,"%] "))',advance='no') &
                achar(13),perc
              percsave = perc
             end if

            end if
            ! On renumerote de 1 à 3 si l'allele 0 est representé dans la population le SNP q
            if ( count(genotype(:,q2) == 0 ) > 0 ) then
                where (genotype(:,q2) >=0 .and. genotype(:,q2) <=2 )
                    Eq2 = genotype(:,q2)+1
                elsewhere
                    Eq2 = 0.d0
                end where
            else
                Eq2 = genotype(:,q2)
                where ( Eq2 > 2 .or. Eq2 < 0)
                    Eq2 = 0
                end where
            end if

            Fq1q2 = dot_product(Eq1,Eq2)

            !print *,Fq1q2
            M(q1,q2) = dot_product(C(q1,:),A(q2,:))*B(q2)
            !print *,M(q1,q2)
            M(q1,q2) = M(q1,q2) + (B(q1)*Fq1q2*B(q2))*invvare
            !print *,"==>",(B(q1)*Fq1q2*B(q2))*invvare,B(q1),B(q2),invvare,Fq1q2,"*"
            M(q1,q2) = M(q1,q2) + dot_product(D(q1,:),Eq2)*B(q2)*invvare
            M(q2,q1) = M(q1,q2)
        !   print *,q1,q2,M(q1,q2)
          !stop
        end do
        M(q1,q1) = B(q1)
     ! print *,q1,q1,M(q1,q1)
    end do
    !$OMP END DO
    deallocate (idmax,ivec,jvec,vvec,Eq1,Eq2)
    !$OMP END PARALLEL

    write (*,*) ""
    !====================================================================================
    !
    ! Decomposition ACP de la matrice Mqq
    !
    !====================================================================================

    ! Etape 4 : Construction de la matrice de transformation

    write (*,*) "Apply an ACP (DSYEV, Matrix dim=",npos,npos,")"
    write (*,*) "======================================"

    allocate (SIGMA(npos))

    JOBZ='V'
    P => M

    call DSYEV_lapack(P,SIGMA,JOBZ)
    print *,"size(Eigenvalues<0) =",count(SIGMA < 0)

    where ( SIGMA < 0.d0 )
        SIGMA = 0.d0
    end where

    allocate (Pdiag(npos))

    ! Calcul de la diagonal de P3 = P * SIGMA * P'
    ! Diag P3 =>
    ! P3 (1,1) = p11*p11*SIGMA(1) + p12*p12*SIGMA(2) + ...

!    allocate (P2(npos,npos))
    ! Calcul de P et Pdiag
    do i=1,npos
        Pdiag(i) = 0
        do j=1,npos
            Pdiag(i) = Pdiag(i) + P(i,j)*P(i,j)*SIGMA(j)
        end do
        P(i,:) = P(i,:)*sqrt(SIGMA)
    end do


    deallocate (SIGMA)
    !====================================================================================
    !
    ! Etape 5 : Simulation des seuils (Methode Muller)
    !
    !====================================================================================
    allocate (Psmin(nsim))
    TAIL_G01ECF = 'U'
    allocate (mts_stream(0:(omp_get_max_threads()*2)),iseed((omp_get_max_threads()*2)))
    call set_mt19937
    call new(mts_stream(0))
    iseed(:((omp_get_max_threads()*2))) = ISEED_CONF(:((omp_get_max_threads()*2)))
    call init(mts_stream(0),iseed)

    print *,"============= SIMUL ===================="
    counti = 0
    !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(Ts,Pvalues,U,isim,i,s,idth)
    allocate (Ts(npos),Pvalues(npos))
    allocate (U(npos))
    Pvalues=0.d0
    idth = (omp_get_thread_num() * 2)+1
    call create_stream(mts_stream(0),mts_stream(idth),idth)
    call create_stream(mts_stream(0),mts_stream(idth+1),idth+1)
    !$OMP DO
    do isim=1,nsim
        !$OMP ATOMIC
        counti=counti+1
        if ( omp_get_thread_num() == 0 ) then
              write (*,fmt='(a,"===>",(" [",i3,"%] "))',advance='no') &
                achar(13),nint((real(counti)/(real(nsim)))*100)
        end if
        do s=1,npos
            u(s) = random_normal(mts_stream(idth),mts_stream(idth+1))
        end do
        !print *,"us:",u
        ! Muq est P...
        u=matmul(P,u)
        ! print *,"as:",u
        Ts=0.d0
        do i=1,npos
            Ts(i)= u(i)*u(i) / Pdiag(i)
            Pvalues(i) = G01ECF(TAIL_G01ECF, Ts(i) , 1.d0, IFAIL)
        end do
        Psmin(isim)=minval(Pvalues)
    end do
    !$OMP END DO
    deallocate (Ts,Pvalues,U)
    !$OMP END PARALLEL
    write (*,fmt='("===>",(" [100%] "))')

    do i=0,omp_get_max_threads()*2
     call delete(mts_stream(i))
    end do

    if ( distributionFile /= "") then

    open(4567,file=trim(chr(ich))//"_"//trim(distributionFile),action="write",iostat=ios)
    if ( ios == 0 ) then
        print *,"======= write ",trim(chr(ich))//"_"//trim(distributionFile)," ====="
        do isim=1,nsim
          write (4567,*) Psmin(isim)
        end do
    end if
    close (4567)
    end if

    write (*,*) "sorting PValues simulation data to get threshold."
    !trie à bulle...
    sort = .true.
    do while ( sort )
        sort = .false.
        do i=1,nsim-1
            if ( Psmin(i) > Psmin(i+1)) then
                vtemp = Psmin(i)
                Psmin(i) = Psmin(i+1)
                Psmin(i+1) = vtemp
                sort = .true.
            end if
        end do
    end do

    write (out_unit,*)
    ! Print the threshold.....
    write (out_unit,*) "-------------------------------------------------"
    write (out_unit,*) "+++++++++++++ REJECTION THRESHOLDS GL=",trim(chr(ich)),"++++++++++++++"
    write (out_unit,*) "-------------------------------------------------"
    write (out_unit,*)
    write (out_unit,fmt="(' ',43('='))")
    write (out_unit,fmt="('|',3x,'Level',2x,'|',2x,'  PValue ',2x,'|',2x,'-Log10(Pvalue)',2x,'|')")
    write (out_unit,fmt="(' ',43('-'))")
    i = nint (( 10 * real(nsim) ) / 100.d0)
    if ( i == 0 ) i = 1
    write (out_unit,fmt="('|',5x,I3,'%',1x,'|',2x,F8.5,3x,'|',3x,F7.3,8x,'|')")&
     10,Psmin(i),-log10(Psmin(i))

    i = nint (( 5 * real(nsim) ) / 100.d0)
    if ( i == 0 ) i = 1
    write (out_unit,fmt="('|',5x,I3,'%',1x,'|',2x,F8.5,3x,'|',3x,F7.3,8x,'|')")&
     5,Psmin(i),-log10(Psmin(i))

    i = nint (( real(nsim) ) / 100.d0)
    if ( i == 0 ) i = 1
    write (out_unit,fmt="('|',5x,I3,'%',1x,'|',2x,F8.5,3x,'|',3x,F7.3,8x,'|')")&
     1,Psmin(i),-log10(Psmin(i))

    write (out_unit,fmt="(' ',43('='))")

    ! Print the biggest pvalues.....
    write (out_unit,*)
    write (out_unit,*) "-------------------------------------------------"
    write (out_unit,*) "++++++++++++++  HIGHEST PVALUES  GL=",trim(chr(ich))," ++++++++++++++++"
    write (out_unit,*) "-------------------------------------------------"

    write (out_unit,*)
    write (out_unit,fmt="(' ',46('='))")
    write (out_unit,fmt="('|',8x,'Rank',2x,'|',2x,'  SNP   ',2x,'|',2x,'-Log10(Pvalue)',2x,'|')")
    write (out_unit,fmt="(' ',46('-'))")

    do i=1,size(maxpvalues,2)
      if ( maxpvalues_index(ich,i) > 0 ) then
       write (out_unit,fmt="('|',5x,I7,2x,'|',2x,I7,3x,'|',3x,F7.3,8x,'|')") &
        i,maxpvalues_index(ich,i),maxpvalues(ich,i)
      end if
    end do

    write (out_unit,fmt="(' ',46('='))")

    write (out_unit,*)
    write (out_unit,*) "------------------------------------------------"
    write (out_unit,*) "++++++++++++++  VISUALIZATION (USING R) GL=",trim(chr(ich))," ++++++++"
    write (out_unit,*) "------------------------------------------------"
    write (out_unit,*)

    write (out_unit,*) "Pvalues distribution (compare theorical distribution with real pvalues) / Manhattan plot"
    call write_scriptR(outFile,"_"//trim(chr(ich))//"_genGraphR.R",chr(ich),solutionFile,npos,nsim,Psmin)

    write (out_unit,*)
    write (out_unit,*) 'run "R -f ',trim(outFile)//"_"//trim(chr(ich))//'_genGraphR.R"',' to generate all graphics.'

    deallocate (Pdiag,P)
    deallocate (Psmin,mts_stream,iseed)

    end do ! nchr

    write (out_unit,*)
    write (out_unit,fmt="(a,a)") 'The GWAS solutions in file :',solutionFile

    deallocate (maxpvalues_index,maxpvalues)

    if ( pvalueNaN > 0 ) then
     write (out_unit,*)
     write (out_unit,*) "------------------------------------------------"
     write (out_unit,*) "+++++ Number of NaN (Not A Number) Pvalue ++++++++"
     write (out_unit,*) "------------------------------------------------"
     write (out_unit,*) pvalueNaN
     write (out_unit,*)
    end if

    write (out_unit,*)
    write (out_unit,*)
    write (out_unit,*) "------------------------------------------------"
    write (out_unit,*) "++++++++++++++++++    END GWAS    ++++++++++++++"
    write (out_unit,*) "------------------------------------------------"
    write (out_unit,*)

    call cpu_time(finish)
    finish = finish - start
    if ( finish < 60 ) then
     write (out_unit,*) "Elapsed time:", nint(finish) ,"sec"
    else
     write (out_unit,*) "Elapsed time:", floor(real(real(finish)/60.0)),"min",nint(modulo(finish,real(60))) ,"sec"
    end if

    deallocate (A,B,C,D)
    deallocate (id,revid_global,genotype)
    deallocate (okmarker,mafinfo)
    deallocate (chr,startpos,endpos,nmark)

contains

    subroutine set_options_gwas

        integer :: nthread

        option_l  = "out"
        option_s_opt = "o"

        if ( exist_arg(option_l,option_s_opt,.true.) ) then
          out_unit=1212
          outFile = get_arg_string(option_l,option_s_opt)
          open(unit=out_unit,file=trim(outFile),iostat=ios,action="write")
          if ( ios /= 0 ) then
           write (0,*) "Can not open file in write mode :",trim(outFile)
           stop
          end if
          write (out_unit,*) "[out file=",trim(outFile),"]"
        else
          ! cette option devient obligatoire
          write (0,*) "no ouput file file defined. Used option --out"
          stop
          out_unit=6 ! standard ouput
          write (out_unit,*) "[out file=<stdout>]"
        end if


        ! Fichier de genotype
        !---------------------------------
        option_l  = "geno"
        option_s_opt = 'g'

        if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
            write (0,*) "* no genotype file defined *"
            stop
        end if

        genotypeFile = get_arg_string(option_l,option_s_opt)
        write (out_unit,*) "[genotype file=",trim(genotypeFile),"]"

        ! Fichier Map
        !---------------------------------
        option_l  = "map"
        option_s_opt = 'c'

        if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
            write (0,*) "* no map file defined *"
            stop
        end if

        mapFile = get_arg_string(option_l,option_s_opt)
        write (out_unit,*) "[map file=",trim(mapFile),"]"

        ! Fichier output Distribution des Psmin
        !---------------------------------
        option_l  = "out-dist"
        option_s_opt = 'y'

        if ( exist_arg(option_l,option_s_opt,.true.) ) then
             distributionFile = get_arg_string(option_l,option_s_opt)
             write (out_unit,*) "[distribution output file=",trim(distributionFile),"]"
        end if


        ! Nombre de caractère que prend l'identifiant dans le fichier de génotypes (10 par défaut)
        !-------------------------------------------------------------------------------------------
        option_l  = "idlength"
        option_s_opt = "i"
        defstr="10"
        sizeid = get_arg_string(option_l,option_s_opt,defstr)
        write (out_unit,*) "[sizeid=",trim(sizeid),"]"

!        ! premiere position testée
!        option_l  = "start"
!        option_s_opt = "a"
!
!        startpos = 1
!
!        if ( exist_arg(option_l,option_s_opt,.true.) ) then
!            startpos = get_arg_int(option_l,option_s_opt)
!        end if
!
!        write (out_unit,*) "[start=",startpos,"]"
!
!        if (startpos<=0) then
!         write (0,*) "prerequisite [startpos>0] "
!         stop
!        end if
!
!        ! derniere position testée
!        option_l  = "end"
!        option_s_opt = "n"
!
!        if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
!            write (0,*) "* no end position to test defined *"
!            stop
!        end if
!
!        endposstring = get_arg_string(option_l,option_s_opt)
!        endpos = get_arg_int(option_l,option_s_opt)
!
!        npostotal = endpos - startpos + 1
!        write (out_unit,*) "[number of pos=",npostotal,"]"

        !
        option_l  = "nmaxpval"
        option_s_opt = "x"

        if ( exist_arg(option_l,option_s_opt,.true.) ) then
            maxpvaluesprint = get_arg_int(option_l,option_s_opt)
        end if

        write (out_unit,*) "[nmaxpval=",maxpvaluesprint,"]"

        ! la variance phenotypique
        option_l     = "vary"
        option_s_opt = "v"
        calcul_vary  =.false.

        if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
            write (out_unit,*) "* no phenotyped variance defined *"
            calcul_vary=.true.
        else
         vary = get_arg_float(option_l,option_s_opt)
         write (out_unit,*) "[vary=",vary,"]"
        end if

        ! l'heritabilité
        option_l  = "h2"
        option_s_opt = "h"

        if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
            write (0,*) "* no heritability defined *"
            stop
        end if

        h2 = get_arg_float(option_l,option_s_opt)
        write (out_unit,*) "[h2=",h2,"]"

         ! MAF
        option_l  = "maf"
        option_s_opt = "m"

        if ( exist_arg(option_l,option_s_opt,.true.) ) then
            minmaf = get_arg_float(option_l,option_s_opt)
        end if

        write (out_unit,*) "[maf=",minmaf,"]"

        if ( minmaf <= 0 ) then
            write (0,*) "Define a MAF strictly positive!"
            stop
        end if

        option_l  = "nsim"
        option_s_opt = "s"

        if ( exist_arg(option_l,option_s_opt,.true.) ) then
           nsim = get_arg_int(option_l,option_s_opt)
        end if

        write (out_unit,*) "[number of simulation=",nsim,"]"

        ! == New ==

        option_l  = "model"
        option_s_opt = "z"

        if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
            write (0,*) "* no model file *"
            stop
        end if

        modelFile = get_arg_string(option_l,option_s_opt)
        write (out_unit,*) "[model file=",trim(modelFile),"]"

        option_l  = "pedig"
        option_s_opt = "p"

        if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
            write (0,*) "* no pedigree file *"
            stop
        end if

        pedigFile = get_arg_string(option_l,option_s_opt)
        write (out_unit,*) "[pedigree file=",trim(pedigFile),"]"

        option_l  = "data"
        option_s_opt = "d"

        if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
            write (0,*) "* no data file *"
            stop
        end if

        dataFile = get_arg_string(option_l,option_s_opt)
        write (out_unit,*) "[data file=",trim(dataFile),"]"

        option_l  = "solution"
        option_s_opt = "b"

        if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
            write (0,*) "* no solution file *"
            stop
        end if

        solutionFile = get_arg_string(option_l,option_s_opt)
        write (out_unit,*) "[solution file=",trim(dataFile),"]"

        option_l  = "nthread"
        option_s_opt = "N"
        nthread=1
        if ( exist_arg(option_l,option_s_opt,.true.) ) then
            nthread = get_arg_int(option_l,option_s_opt)
        end if

        if ( nthread < 1 ) then
         write (0,*) "Thread number have to greater or equal than 1! [",nthread,"]"
         stop
        end if

        write (out_unit,*) "[nthread=",nthread,"]"
        call OMP_SET_NUM_THREADS(nthread)
#ifdef MKL_INTEL
        call MKL_SET_NUM_THREADS(nthread)
#endif
        if ( size (ISEED_CONF) < omp_get_num_threads()*2) then
         write (0,*) "To much SEED required by the mersenne twister random generator number. Decrease the number of thread !"
         stop
        end if

        write (out_unit,*)
        write (out_unit,*) "==========================================="
        write (out_unit,*)

    end  subroutine set_options_gwas


    !> Cette routine construit les parametres des donnees (performance et effets de nuisances) necessaires à l'execution de blupf90.
    !> Ces parametres sont décrit par l'utilsiateur.
    !>
    !> @param dataFile The logical expression to test. (Explanation of individual arguments)
    !> @param modelFile The message to print on fail. If omitted, a generic message is printed.
    !> @param paramblupf90 Controls whether to stop execution. If doStop .true.,
    !> @see type(Param_blupf90)
    !> @param ndata number of data record (each line corresponding to an animal)
    !> @param ncol number of data column (phenotype, weight, SNP statut, nuisance effect )
    !> @param data array of data (two dimension)
    !> @author Olivier Filangi
    !> @version 1.0 (Version information)
    !> @warning L'effet animal est ajouté artificiellement...c'est le dernier effet du model voir fill_structure_pedigree_blupf90

    subroutine fill_structure_data_blupf90(dataFile,modelFile,paramblupf90,ndata,ncol,data)
        character(len=len_long_cli)  , intent(in)  :: modelFile    ! fichier de descripiton du model
        character(len=len_long_cli)  , intent(in)  :: dataFile     ! fichier des donnees (performance+effet de nuissance)
        type(Param_blupf90), intent(inout)         :: paramblupf90 ! Strucutre de donnée pour l'utilisation de la routine de blupf90
        integer ,intent(inout)                     :: ndata,ncol
        real, dimension(:,:),pointer,intent(inout) :: data
        !local
        integer :: unit_val=1234,ios,lev,pos,pos_id_record,maxfield,pos_weight,pos_y
        character(len=50) :: section,lexeme
        integer,parameter::   maxcorr=50   ! maximum number of correlated effects; used for g 50 / Pris de blupf90
        real, dimension(:,:),allocatable :: datauser
        logical :: neff_set,nrecord_set,observation_set

        print *,"read model file..."

        neff_set=.false.
        nrecord_set=.false.
        observation_set=.false.
        ! fixe, analyse unicaractere

        paramblupf90%ntrait = 1
        !allocation depending ntrait
        allocate (paramblupf90%pos_y(paramblupf90%ntrait))
        allocate (paramblupf90%r(paramblupf90%ntrait,paramblupf90%ntrait))
        !end allocation depending ntrait

        !par defaut la quatrieme colonne est le poid de la performance dans le fichier generer
        paramblupf90%pos_weight=4
        ! la cinquieme colonne est la performance
        paramblupf90%pos_y(1)=5
        !position des poids des phenos par defaut dans le fichier data
        pos_weight=0

        !position des ID animaux par defaut dans le fichier data
        pos_id_record=1

        open(unit=unit_val,file=modelFile,iostat=ios,action="read")

        if ( ios /= 0 ) then
            write (0,*) "Can not open file ",trim(modelFile)
            stop
        end if

        paramblupf90%neff = 0

        do while (ios == 0 )
            read (unit_val,'(a)',iostat=ios) section
            if ( ios /= 0 ) cycle
            if ( trim(section) == "" ) cycle
            section=adjustl(section)
            if ( section == "NUMBER_OF_EFFECTS" ) then
                read (unit_val,'(I7)',iostat=ios), paramblupf90%neff
                if ( ios /= 0 ) then
                    write (0,*) "Can read number of effect in the model file ",trim(modelFile)
                    stop
                end if

                ! On ajoute 3 effets qu'on gere en interne
                ! le premier effet est la moyenne general
                ! l'avant dernier effet est l'effet SNP
                ! le dernier effet est l'effet polygenique

                paramblupf90%neff = paramblupf90%neff + 3

                !allocation depending ntrait/neffect
                allocate (paramblupf90%pos_eff(paramblupf90%neff,paramblupf90%ntrait))
                allocate (paramblupf90%nlev(paramblupf90%neff))
                allocate (paramblupf90%effecttype(paramblupf90%neff))
                allocate (paramblupf90%nestedcov(paramblupf90%neff,paramblupf90%ntrait))
                allocate (paramblupf90%randomnumb(paramblupf90%neff))
                allocate (paramblupf90%randomtype(paramblupf90%neff))
                allocate (paramblupf90%gcov(paramblupf90%neff,maxcorr,maxcorr))
               !end allocation depending ntrait/neffect

                neff_set=.true.

            else if ( section(1:11) == "OBSERVATION") then
                read (unit_val,'(I7)',iostat=ios), pos_y
                if ( ios /= 0 ) then
                    write (0,*) "Can read position of trait in the model file ",trim(modelFile)
                    stop
                end if
                observation_set = .true.
            else if ( section(1:6) == "WEIGHT") then
                lexeme=""
                read (unit_val,*,iostat=ios), lexeme
                lexeme=adjustl(lexeme)
                read(lexeme,'(I7)',iostat=ios) pos_weight
                if ( ios /= 0 ) then
                    write (0,*) "Can read weight trait in the model file ",trim(modelFile)
                    stop
                end if

            else if ( section(1:9) == "RECORD ID") then
                read (unit_val,'(I7)',iostat=ios), pos_id_record
                if ( ios /= 0 ) then
                    write (0,*) "Can read position of record id animal in the model file ",trim(modelFile)
                    stop
                end if
                nrecord_set = .true.
            else if (section(1:9) == "EFFECTS:") then
                if ( paramblupf90%neff == 0 ) then
                    write (0,*) "Defined the section NUMBER_OF_EFFECT in the model file"
                    stop
                end if
                do i=1,paramblupf90%neff-3
                    read (unit_val,*,iostat=ios) pos,lev,lexeme
                    if ( ios/=0 ) then
                        write (0,*) "Can not read the ",i," effect, EFFECTS section, in the model file ",trim(modelFile)
                        stop
                    end if
                    lexeme=adjustl(lexeme)
                    paramblupf90%pos_eff(i+1,:)=pos ! attention position dans le fichier de performance initiale...
                    paramblupf90%nlev(i+1)=lev
                    if ( lexeme(1:5) == "cross") then
                        paramblupf90%effecttype(i+1)=effcross
                    else if (lexeme(1:3) == "cov") then
                        paramblupf90%effecttype(i+1)=effcov
                        if ( lev /= 1 ) then
                         write (0,*) "nlev covariate must to be equal to 1. "
                         write (0,*) "===>",pos,lev,lexeme
                         stop
                        end if
                    else
                        write (0,*) i," effect=> Can not manage other effect than cov/cross !"
                        stop
                    end if
                end do
            else
                print *,"unknown section:",trim(section)
            end if
        end do

        close (unit_val)

        if ( .not. neff_set ) then
          write (0,*) "section NUMBER_OF_EFFECTS is not defined !"
          stop
        end if

        if ( .not. nrecord_set ) then
          write (0,*) "section RECORD ID is not defined !"
          stop
        end if

        if ( .not. observation_set ) then
          write (0,*) "section OBSERVATION is not defined !"
          stop
        end if

        !position of nesting effect for each trait
        paramblupf90%nestedcov = 0

        !moyenne generale
        paramblupf90%pos_eff(1,:)=2
        paramblupf90%nlev(1)=1
        paramblupf90%effecttype(1)=effcross

        !SNP
        paramblupf90%pos_eff(paramblupf90%neff-1,:)=3
        paramblupf90%nlev(paramblupf90%neff-1)=1 ! à changer pour le traitement de chaque SNP
        paramblupf90%effecttype(paramblupf90%neff-1)=effcov

        !Effet polygenique est initialisé dans fill_structure_pedigree_blupf90

        !recherche du maximum de nombre de champ à lire dans le fichier data de l'utilisateur
        maxfield=0

        do i=1,paramblupf90%neff-3
            maxfield=max(maxfield,paramblupf90%pos_eff(i+1,1))
        end do

        maxfield=max(maxfield,pos_weight)
        maxfield=max(maxfield,pos_id_record)
        maxfield=max(maxfield,pos_id_record)
        maxfield=max(maxfield,pos_y)

        ndata = 0
        open(unit=unit_val,file=dataFile,iostat=ios,action="read")
        if ( ios /= 0 ) then
            write (0,*) "Can not open file ",trim(dataFile)
            stop
        end if

        do while ( ios == 0 )
            read (unit_val,'(a)',iostat=ios) lexeme
            if ( ios == 0 ) ndata = ndata + 1
        end do
        write (*,*) "[data] number of record =",ndata
        rewind(unit_val)

        if ( ndata == 0 ) then
            write (0,*) "Can not read data user."
            stop
        end if

        if ( associated(data) ) then
            write (0,*) "Devel error. data array is ever allocated !"
            stop
        end if

        allocate (datauser(ndata+1,maxfield))
        ncol = 5+paramblupf90%neff-3
        allocate (data(ndata,ncol))

        ios = 0
        i=1
        write (lexeme,fmt="(I7)") maxfield
        do while ( ios == 0 )
            read (unit_val,*,iostat=ios) datauser(i,:)
            if ( ios == 0 ) i = i + 1
        end do

        close (unit_val)

        ! Construction du tableau data
        ! ID => 1ere colonne
        data(:,1) = datauser(:ndata,pos_id_record)
        ! Effet moyenne generale => 2eme colonne
        data(:,2) = 1.d0
        ! Effet snp
        data(:,3)=0 ! bidon une routine doit modifier cette colonne pour chaque position testée
        ! Poid => 4ieme colonne

        if ( pos_weight /= 0 ) then
            data(:,4) = datauser(:ndata,pos_weight)
        else
            data(:,4) = 1.d0
        end if
        ! Performance => 5ieme colonne
        data(:,5) = datauser(:ndata,pos_y)
        !         print *,datauser(:ndata,3)
        !         stop
        ! Puis les effets de nuisance
        ! Attention, on reinitialise la position de ces effets par rapport au tableau data
        do i=1,paramblupf90%neff-3
            data(:,5+i) = datauser(:ndata,paramblupf90%pos_eff(i+1,1))
            paramblupf90%pos_eff(i+1,1) = 5+i
        end do

        deallocate (datauser)
    end subroutine fill_structure_data_blupf90

    !> Cette routine construit les parametres des donnees de genealogie necessaires à l'execution de blupf90.
    !>
    !> @param pedigFile chemin du fichier de genealogie donné par l'utilisateur
    !> @param paramblupf90 structure d'entrée utilisable par la routine blupf90,
    !> @see type(Param_blupf90)
    !> @param ndata_ped argument initialisé en sortie de procedure. correspond au nombre d'enregistrement du fichier de genealogie (1ere dimension du parametre data_ped)
    !> @param data_ped Information de la généalogie. tableau à deux dimension comportant 3 colonnes et le nombre d'"enregistrement lu" lignes. Ce tableau est alloué dans cette routine.
    !> @author Olivier Filangi
    !> @version 1.0 (Version information)
    !> @warning Une erreur est remonté sur le tableau data_ped est initalisé.

    subroutine fill_structure_pedigree_blupf90(pedigFile,paramblupf90,ndata_ped,data_ped)
        character(len=len_long_cli)  , intent(in)  :: pedigFile    ! pedigree
        type(Param_blupf90), intent(inout)         :: paramblupf90 ! Structure de donnée pour l'utilisation de la routine de blupf90
        integer            , intent(inout)         :: ndata_ped    ! nombre d'entrée du fichier de genealogy conditionnant la taille du tableau data_ped
        integer, pointer, dimension(:,:), intent(inout) :: data_ped ! tableau à 3 colonne contenant tous les enregistrement du fichier de genealogy
        integer :: ios,id
        integer :: unit_val=5432,idan,pere,mere

        character(len=20) :: section

        if ( associated(data_ped) ) then
            write (0,*) "Devel Error. data_ped is initialized."
            stop
        end if

        open(unit=unit_val,file=pedigFile,iostat=ios,action="read")
        if ( ios /= 0 ) then
            write (0,*) "Can not open file ",trim(pedigFile)
            stop
        end if

        !Comptage des enregistrements pour l'allocation de data_ped
        ndata_ped=0
        do while ( ios == 0 )
            read (unit_val,*,iostat=ios) section
            if ( ios == 0 ) ndata_ped=ndata_ped+1
        end do

        rewind(unit_val)
        write (*,*) "[pedigree] Number of record=",ndata_ped

        if (ndata_ped == 0 ) then
            write (0,*) "Can not read pedigree file."
            stop
        end if

        allocate (data_ped(ndata_ped,3))
        ios= 0
        id=0
        do while ( ios == 0 )
            read (unit_val,*,iostat=ios) idan,pere,mere
            if ( ios == 0 ) then
                id=id+1
                data_ped(id,1) = idan
                data_ped(id,2) = pere
                data_ped(id,3) = mere
            end if
        end do

        close (unit_val)

        ! Il faut remplir le dernier effet : effet polygenique

        paramblupf90%pos_eff(paramblupf90%neff,:)=1 ! c'est la premiere colonne du fichier de performance....
        paramblupf90%nlev(paramblupf90%neff)=data_ped(ndata_ped,1) ! nombre d animaux...dernier animal si genealogy bien triée
        paramblupf90%effecttype(paramblupf90%neff)=effcross ! en effet fixé

        paramblupf90%randomtype=g_fixed
        paramblupf90%randomnumb=0
        !L'effet 5
        paramblupf90%randomnumb(paramblupf90%neff)=1
        paramblupf90%randomtype(paramblupf90%neff)=g_A !add_animal

    end subroutine fill_structure_pedigree_blupf90

    !> Cette routine initialise la variance genetique pour l'execution de blupf90.
    !>
    !> @param paramblupf90 structure d'entrée utilisable par la routine blupf90,
    !> @see type(Param_blupf90)
    !> @param covgen valeur de variance genetique pour le caractere d'etude
    subroutine set_covariance_genetique_blupf90(paramblupf90,covgen)
        type(Param_blupf90), intent(inout) :: paramblupf90 ! Structure de donnée pour l'utilisation de la routine de blupf90
        real(kind=dp)      ,intent(in)     :: covgen
        paramblupf90%gcov(paramblupf90%neff,1,1)=covgen
    end subroutine set_covariance_genetique_blupf90

    !> Cette routine initialise la variance residuelle pour l'execution de blupf90.
    !>
    !> @param paramblupf90 structure d'entrée utilisable par la routine blupf90,
    !> @see type(Param_blupf90)
    !> @param covgen valeur de variance genetique pour le caractere d'etude
    subroutine set_residuelle_blupf90(paramblupf90,res)
        type(Param_blupf90), intent(inout)  :: paramblupf90 ! Structure de donnée pour l'utilisation de la routine de blupf90
        real(kind=dp)      , intent(in)     :: res
        paramblupf90%r(1,1)=res
    end subroutine set_residuelle_blupf90

    !> Cette routine initialise les parametres generaux de l'execution de blupf90.
    !>
    !> @param paramblupf90 structure d'entrée utilisable par la routine blupf90,
    !> @see type(Param_blupf90)
    subroutine set_general_blupf90(paramblupf90)
        type(Param_blupf90), intent(inout) :: paramblupf90 ! Structure de donnée pour l'utilisation de la routine de blupf90
        integer :: i

        paramblupf90%solv_method="FSPAK"

        allocate(paramblupf90%offset(paramblupf90%neff))
        do i=1,paramblupf90%neff
            paramblupf90%offset(i)=sum(paramblupf90%nlev(1:i-1))*paramblupf90%ntrait
        enddo

    end subroutine set_general_blupf90


    !> Cette routine affiche les informations des parametres pour executer un blupf90.
    !>
    !> @param paramblupf90 structure d'entrée utilisable par la routine blupf90,
    !> @see type(Param_blupf90)
    subroutine print_information_model_blupf90(paramblupf90)
        type(Param_blupf90), intent(inout) :: paramblupf90 ! Structure de donnée pour l'utilisation de la routine de blupf90

        print *,"=== [ param blupf90 ] ==="

        print *,"[number of trait]        =",paramblupf90%ntrait
        print *,"[number of effects]      =",paramblupf90%neff
        print *,"[pos y in data]          =",paramblupf90%pos_y
        print *,"[pos weight in data]     =",paramblupf90%pos_weight
        do i=1,paramblupf90%neff
            print *,"[effect]  =",i
            print *,i," [pos eff]    =",paramblupf90%pos_eff(i,:)
            print *,i," [nlev]       =",paramblupf90%nlev(i)
            print *,i," [type]       =",paramblupf90%effecttype(i)
            print *,i," [nested]     =",paramblupf90%nestedcov(i,:)
            print *,i," [randomtype] =",paramblupf90%randomtype(i)
            print *,i," [randomnumb] =",paramblupf90%randomnumb(i)
            print *,i," [random covariance matrix] =",paramblupf90%gcov(i,1,1)
        end do
        print *,"[residual]               =",paramblupf90%r(1,1)
        print *,""
        print *," === parameter === "
        print *,"[con crit]               =",paramblupf90%conv_crit
        print *,"[maxrounds]              =",paramblupf90%maxrounds
        print *,"[solv_method]            =",paramblupf90%solv_method
        print *,"[r_factor]               =",paramblupf90%r_factor
        print *,"[blksize]                =",paramblupf90%blksize
        print *,"[mis]                    =",paramblupf90%mis
        print *,"[lchk_lev]               =",paramblupf90%lchk_lev

    end subroutine print_information_model_blupf90

    !> Cette routine modifie la colonne du status du SNP dans le tableau data.(prealable pour une analyse GWAS)
    !> on crée un nouveau tableau car il peut y avoir des donnees manquantes
    !> @param nu number of anaimal genotyped
    !> @param revid array to get genotype id animal
    !> @param genotyped array of genotyped animals
    !> @param npos number of marker
    !> @param nummark marker to set in the nex data
    !> @param ndata number of record of original data
    !> @param ncol number of column of orginal data
    !> @param newdata number of record of the data to tested with the blup at the marker position
    !> @param nnewdata new data

    subroutine set_marker_blupf90(revid,nu,npos,genotype,ndata,ncol,data,nnewdata,newdata,nummark,revid_currenthaplo)
        integer  ,dimension(:)     ,intent(in)     :: revid
        integer                    ,intent(in)     :: nu
        integer                    ,intent(in)     :: npos
        integer ,dimension(nu,npos),intent(inout)  :: genotype
        integer                    ,intent(in)     :: ndata
        integer                    ,intent(in)     :: ncol
        real, dimension(ndata,ncol),intent(in)     :: data
        integer                    ,intent(inout)  :: nnewdata
        real, dimension(ndata,ncol),intent(inout)  :: newdata
        integer                    ,intent(in)     :: nummark
        integer  ,dimension(ndata) ,intent(inout)  :: revid_currenthaplo ! correspondance entre la liste data (phenotype)/genotypes


        integer :: i,ida,sizerevid

        if ( nummark <= 0 .or. nummark > size(genotype,2)) then
            write (0,*) "[Devel error] bad init of numark!"
            stop
        end if

        sizerevid = size(revid)
        nnewdata=0
        revid_currenthaplo=0

        do i=1,ndata
            ida = nint(data(i,1)) ! ID animal
            if ( ida > sizerevid ) cycle ! none corresponding
            if ( revid(ida) == -1 ) cycle ! none corresponding
            if ( genotype(revid(ida),nummark) == 5 ) cycle ! genotype missing data
            nnewdata=nnewdata+1
            newdata(nnewdata,:)=data(i,:)
            !status au QTL
            newdata(nnewdata,3) = real(genotype(revid(ida),nummark))
          !revid_currenthaplo(i)=revid(ida)
          !print *,ida,revid(ida),nummark,genotype(revid(ida),:)
        end do

        if ( nnewdata == 0 ) then
         write (0,*) "All ID animal from data do not match with ID animal in genotype file !"
         stop
        end if

        !Securite...a voir apres si on enleve
        if (any(newdata(:nnewdata,3) < 0 ) .or. any(newdata(:nnewdata,3) > 5 ) ) then
            print *,"========================DATA ======================"
            print *,newdata(:nnewdata,3)
            print *
            print *
            print *,"========================GENOTYPES==================="
            print *,genotype(:,nummark)

            write (0,*) "Error dev==>set_marker_blupf90 bad init marker 1!"
            stop
        end if

        ! On autorise pas la valeur 0 dans les niveaux des effets SNP
        if (any(newdata(:nnewdata,3) == 0 )) then
            newdata(:nnewdata,3) = newdata(:nnewdata,3) + 1
        end if

    end subroutine set_marker_blupf90

    subroutine remove_maf(nu,start,npos,genotype,okmarker,minmaf,mafinfo)
        integer            , intent(in) :: nu ! number of animal
        integer            , intent(in) :: start ! first marker to treat
        integer            , intent(in) :: npos ! number of marker from genotype file
        integer , dimension(nu,npos) ,intent(in)  :: genotype
        logical ,dimension(start:npos),intent(out) :: okmarker
        real(kind=dp)           ,intent(in)  :: minmaf
        real , dimension(start:npos)  ,intent(out) :: mafinfo
        real ,dimension(nu) :: buf
        integer :: i

        okmarker = .false.
        mafinfo = 0.0

        do ich=1,nchr
         do i=startpos(ich),endpos(ich)
            if ( mafinfo(i) == 0.0  ) then
             buf = genotype(:,i)
             where (buf == 5) buf = 0
             buf = buf/2
             !comtage du éieme allele
             mafinfo(i)  = sum(buf) / real(nu) ;
             okmarker(i) = (mafinfo(i)>minmaf) .and. (((1-mafinfo(i))>minmaf))
            end if
         end do
        end do

    end subroutine remove_maf

    subroutine fill_matrix_snp(nu,nb,idmat_vect,j_vect,values_vect,A,B,C,D)
        integer            , intent(in) :: nu ! number of animal
        integer            , intent(in) :: nb ! number of level for all fixed effect
        integer            , intent(in) :: idmat_vect(nb+1+nb+nu+1)
        integer            , intent(in) :: j_vect(nb+1+nb+nu+1)
        real(kind=dp)      , intent(in) :: values_vect(nb+1+nb+nu+1)
        real(kind=dp)      , intent(inout) , dimension(nb) :: A
        real(kind=dp)      , intent(inout)                 :: B
        real(kind=dp)      , intent(inout) , dimension(nb) :: C
        real(kind=dp)      , intent(inout) , dimension(nu) :: D

        integer :: ndim,i,centries

        ndim = nb+1+nb+nu

        do i=1,ndim
            select case (idmat_vect(i))
                case (0)
                    A(j_vect(i)) = values_vect(i)
                case (1)
                    B = values_vect(i)
                case (2)
                    C(j_vect(i)) = values_vect(i)
                case (3)
                    D(revid_global(j_vect(i))) = values_vect(i)
                case default

            end select
        end do

    end subroutine fill_matrix_snp

    subroutine fill_matrix_haplotype(nu,nb,nhap,idmat_vect,i_vect,j_vect,values_vect,A,B,C,D)
        integer            , intent(in) :: nu ! number of animal
        integer            , intent(in) :: nb ! number of level for all fixed effect
        integer            , intent(in) :: nhap ! number of level for the Haplotype effect (==1 for SNP)
        integer            , intent(in) :: idmat_vect(nb+1+nb+nu+1)
        integer            , intent(in) :: i_vect(nb+1+nb+nu+1)
        integer            , intent(in) :: j_vect(nb+1+nb+nu+1)
        real(kind=dp)      , intent(in) :: values_vect(nb+1+nb+nu+1)
        real(kind=dp)      , intent(inout) , dimension(nhap,nb)   :: A
        real(kind=dp)      , intent(inout) , dimension(nhap,nhap) :: B
        real(kind=dp)      , intent(inout) , dimension(nhap,nb)   :: C
        real(kind=dp)      , intent(inout) , dimension(nhap,nu)   :: D

    end subroutine fill_matrix_haplotype


    subroutine init_random()
        integer :: n,clock,i
        integer, dimension(:), allocatable :: seed

        !initialisation pour les routines intrasec de fortran
        call random_seed(size=n)
        allocate(seed(n))
        call system_clock(count=clock)
        seed = clock + 37 * (/ (i - 1, i = 1, n) /)
        call random_seed(put = seed)
        deallocate(seed)

    end subroutine init_random

    subroutine rand_number(u)
        real , intent(out) :: u
        call random_number(u)
    end subroutine rand_number

    function normal(mean,sigma) !returns a normal distribution
        real(kind=dp)  , intent(in) :: mean,sigma
        real(kind=dp) normal,tmp
        integer flag
        real(kind=dp) fac,gsave,rsq,r1,r2,u
        save flag,gsave
        data flag /0/
        !$omp threadprivate (flag)
        if (flag.eq.0) then
            rsq=2.0
            do while(rsq.ge.1.0.or.rsq.eq.0.0) ! new from for do
                call random_number(u)
                r1=2.0*u-1.0
                call random_number(u)
                r2=2.0*u-1.0
                rsq=r1*r1+r2*r2
            enddo
            fac=sqrt(-2.0*log(rsq)/rsq)
            gsave=r1*fac
            tmp=r2*fac
            flag=1
        else
            tmp=gsave
            flag=0
        endif
        normal=tmp*sigma+mean
        return
    end function normal


    FUNCTION random_normal(mts1,mts2) RESULT(fn_val)

        ! Adapted from the following Fortran 77 code
        !      ALGORITHM 712, COLLECTED ALGORITHMS FROM ACM.
        !      THIS WORK PUBLISHED IN TRANSACTIONS ON MATHEMATICAL SOFTWARE,
        !      VOL. 18, NO. 4, DECEMBER, 1992, PP. 434-435.

        !  The function random_normal() returns a normally distributed pseudo-random
        !  number with zero mean and unit variance.

        !  The algorithm uses the ratio of uniforms method of A.J. Kinderman
        !  and J.F. Monahan augmented with quadratic bounding curves.
        type(mt_state) :: mts1,mts2
        REAL(kind=dp)  :: fn_val

        !     Local variables
        REAL(kind=dp)     :: s = 0.449871, t = -0.386595, a = 0.19600, b = 0.25472,    &
            r1 = 0.27597, r2 = 0.27846, u, v, x, y, q

        !     Generate P = (u,v) uniform in rectangle enclosing acceptance region

        DO
!            CALL RANDOM_NUMBER(u)
!            CALL RANDOM_NUMBER(v)
            u = genrand_double1(mts1)
            v = genrand_double1(mts2)
            v = 1.7156 * (v - half)

            !     Evaluate the quadratic form
            x = u - s
            y = ABS(v) - t
            q = x**2 + y*(a*y - b*x)

            !     Accept P if inside inner ellipse
            IF (q < r1) EXIT
            !     Reject P if outside outer ellipse
            IF (q > r2) CYCLE
            !     Reject P if outside acceptance region
            IF (v**2 < -4.0*LOG(u)*u**2) EXIT
        END DO

        !     Return ratio of P's coordinates as the normal deviate
        fn_val = v/u
        RETURN

    END FUNCTION random_normal

    function G01ECF(TAIL, X, DF, IFAIL) result(res)
        integer , intent(out)        :: IFAIL
        real(kind=dp),intent(in)     :: X, DF
        character(len=1) ,intent(in) :: TAIL

        integer                      :: ind
        real                         :: Xreal,res
        real, external :: prob

        ind=int(DF)
        Xreal=X
        RES=prob(Xreal,ind)

        IF ( TAIL == 'U' ) RETURN
        IF ( TAIL == 'L') RES = 1-RES

    end function G01ECF

end program gwas_blupf90_muller
