module out_scriptR
    use kinds

    implicit none

    character(len=20),parameter  ,private           :: colMuller5p      =   'rgb(0.9,0.1,0)'
    character(len=20),parameter  ,private           :: colMuller10p     =   'rgb(0.7,0.3,0)'
    character(len=20),parameter  ,private           :: colBonferroni5p  =   'rgb(0.3,0,0.7)'
    character(len=20),parameter  ,private           :: colBonferroni10p =   'rgb(0.1,0,0.9)'

    public  :: write_scriptR
    private :: write_pvalue,write_manhattan_plot

contains

    subroutine write_scriptR(fileout,filename,chromo,solutionFile,npos,nsim,Psmin)
        character(len=*) ,intent(in)  :: fileout
        character(len=*) ,intent(in)  :: filename
        character(len=*) ,intent(in)  :: chromo
        character(len=*) ,intent(in)  :: solutionFile
        integer          ,intent(in)  :: npos
        integer          ,intent(in)  :: nsim
        real(kind=dp)    , dimension(nsim) ,intent(in) :: Psmin

        character(len=5000) :: fileo
        integer :: unit_R = 3456
        integer :: ios

        fileo = trim(fileout)//trim(filename)

        open(unit=unit_R,file=fileo,iostat=ios)

        if ( ios /= 0 ) then
            write (0,*) "Can not write file : ",trim(fileo)
            return
        end if

        write (unit_R,*) '#!/usr/bin/Rscript'
        call write_pvalue(unit_R,solutionFile,chromo,npos)
        call write_manhattan_plot(unit_R,solutionFile,chromo,npos,nsim,Psmin)
        close(unit_R)

    end subroutine write_scriptR

    subroutine write_pvalue(unit_R,solutionFile,chromo,npos)
        integer, intent(in) :: unit_R
        character(len=*) ,intent(in)  :: solutionFile
        character(len=*) ,intent(in)  :: chromo
        integer          ,intent(in)  :: npos

        ! COMPARAISON DISTRIB H0
        write (unit_R,*) "png(file='"//trim(chromo)//"_"//trim(solutionFile)//"_pvalues.png',width=800,res=120)"
        write (unit_R,*) "snps=read.table('"//trim(solutionFile)//"',header=T)"
        write (unit_R,*) 'yy<-sort(-log10(snps$pvalue[snps$chr=="'//trim(chromo)//'"]))'
        write (unit_R,*) "xx<-sort(-log10((1:length(yy))/length(yy)))"
        write (unit_R,*) 'plot(xx,yy,xlab="pvalues GWAS",ylab="pvalues théorique")'
        write (unit_R,*) 'abline(a=0,b=1,col="red")'
        write (unit_R,fmt='(a)') 'title(main="p_values obervées en fonction des théoriques '//&
        trim(chromo)//'", col.main="blue", font.main=2)'

    end subroutine write_pvalue

    subroutine write_manhattan_plot(unit_R,solutionFile,chromo,npos,nsim,Psmin)
        integer, intent(in) :: unit_R
        character(len=*) ,intent(in)  :: solutionFile
        character(len=*) ,intent(in)  :: chromo
        integer          ,intent(in)  :: npos
        integer          ,intent(in)  :: nsim
        real(kind=dp)    , dimension(nsim) ,intent(in) :: Psmin


        integer :: i

        !!! MANHATTAN PLOT
        write (unit_R,*) "png(file='"//trim(chromo)//"_"//trim(solutionFile)//"_manhattan_plot.png',width=800,res=120)"
        !i = nint (( real(nsim) ) / 100.d0)
        i = 1
        write (unit_R,*) 'yy<-(-log10(snps$pvalue[snps$chr=="'//trim(chromo)//'"]))'
        write (unit_R,*) 'plot(snps$snp[snps$chr=="'//trim(chromo)//'"],yy,col="green",ylim=c(0,-log10(',Psmin(i),'))',&
            ',xlab="positions",ylab="-log10(pvalues)")'
        write (unit_R,fmt='(a)') 'title(main="Manhattan plot - Seuils Muller/Bonferroni '//&
        trim(chromo)//'", col.main="blue", font.main=2)'
        i = nint (( 5 * real(nsim) ) / 100.d0)
        write (unit_R,*) 'abline(h=-log10(',Psmin(i),'),col=',colMuller5p,',lty=3)' ! seuil muller 5%
        i = nint (( 10 * real(nsim) ) / 100.d0)
        write (unit_R,*) 'abline(h=-log10(',Psmin(i),'),col=',colMuller10p,',lty=3)' ! seuil muller 10%
        write (unit_R,*) "ll<-length(sort(yy))"
        write (unit_R,*) 'abline(h=-log10(0.10/ll),col=',colBonferroni10p,',lty=2)'! seuil bonferroni 10%
        write (unit_R,*) 'abline(h=-log10(0.05/ll),col=',colBonferroni5p,',lty=2)'! seuil bonferroni 5%
        write (unit_R,fmt='(a)') 'legend(x="topleft",cex=0.6,c("Bonferroni 5%", "Bonferroni 10%", "Muller 5%", "Muller 10%"),',&
            'col=c('//trim(colBonferroni5p)//','//trim(colBonferroni10p)//','//trim(colMuller5p)//&
            ','//trim(colMuller10p)//'),lwd=c(1,1,1,1),lty=c(2,2,3,3))'
        write (unit_R,*) 'paste("Number of VALID SNP =",ll," / ',npos,'",sep=" ")'

    end subroutine write_manhattan_plot


end module out_scriptR
